
#include <exec/types.h>

#include <dos/dos.h>
#include <dos/dosextens.h>
#include <dos/dosasl.h>

#include <proto/dos.h>

BOOL SoftlinkDODIR(struct AnchorPath *ap,
                   LONG followsoft,
                   LONG quiet,
                   struct DosLibrary *DOSBase)
{
    BOOL enter;
    BPTR dirlock, lock;
    UBYTE tmpbuf[512];

    if (ap->ap_Info.fib_DirEntryType < 0)
    {
        /* It's a file, don't enter */
        return FALSE;
    }
    else if (ap->ap_Info.fib_DirEntryType != ST_SOFTLINK)
    {
        /* Not a softlink, do enter */
        return TRUE;
    }

    /* Ok, it's a softlink, take a closer look... */

    if (!followsoft)
    {
        /* Oops, caller didn't want us to enter it! */
        return FALSE;
    }

    enter = TRUE;

    dirlock = CurrentDir(ap->ap_Current->an_Lock);
    lock = Lock(ap->ap_Info.fib_FileName, ACCESS_READ);
    if (lock)
    {
        struct FileInfoBlock *fib = (APTR) (((ULONG) tmpbuf + 3) & ~3);
        LONG rc;
        /* NOTE: For example Search depends on ap_Info.fib_Size64 from MatchNext */
        if (LIB_MINVER(&DOSBase->dl_lib, 51, 28))
            rc = Examine64(lock, fib, NULL);
        else
            rc = Examine(lock, fib);
        if (rc)
        {
            if (fib->fib_DirEntryType < 0)
            {
                /* Softlink to a file, don't enter */
                enter = FALSE;
            }

            /* copy everything but the fib_FileName */
            ap->ap_Info.fib_DiskKey      = fib->fib_DiskKey;
            ap->ap_Info.fib_DirEntryType = fib->fib_DirEntryType;
            ap->ap_Info.fib_Protection   = fib->fib_Protection;
            ap->ap_Info.fib_EntryType    = fib->fib_EntryType;
            ap->ap_Info.fib_Size         = fib->fib_Size;
            ap->ap_Info.fib_NumBlocks    = fib->fib_NumBlocks;
            ap->ap_Info.fib_Date         = fib->fib_Date;
            strcpy(ap->ap_Info.fib_Comment, fib->fib_Comment);
            ap->ap_Info.fib_OwnerUID     = fib->fib_OwnerUID;
            ap->ap_Info.fib_OwnerGID     = fib->fib_OwnerGID;
            ap->ap_Info.fib_Size64       = fib->fib_Size64;
            ap->ap_Info.fib_NumBlocks64  = fib->fib_NumBlocks64;
            memcpy(ap->ap_Info.fib_Reserved, fib->fib_Reserved, sizeof(fib->fib_Reserved));
        }

        UnLock(lock);
    }
    else
    {
        struct DevProc *dvp;
        LONG ioerr = IoErr();

        if (ioerr == ERROR_OBJECT_NOT_FOUND &&
            (dvp = GetDeviceProc("", NULL)))
        {
            if (ReadLink(dvp->dvp_Port, dvp->dvp_Lock, ap->ap_Info.fib_FileName, tmpbuf, sizeof(tmpbuf) - 1) > 0)
            {
                if (!quiet)
                {
                    tmpbuf[sizeof(tmpbuf) - 1] = '\0';

                    Printf("Warning: Skipping dangling softlink %s -> %s\n",
                           (LONG) ap->ap_Info.fib_FileName, (LONG) tmpbuf);
                }

                /* Dangling softlink, don't enter */
                enter = FALSE;
            }

            FreeDeviceProc(dvp);
        }

        SetIoErr(ioerr);
    }
    CurrentDir(dirlock);

    return enter;
}
